from typing import Any

import neptune_scale.api.run
from neptune_scale.api.attribute import Attribute

__all__ = ("Run",)


class Run(neptune_scale.api.run.Run):
    """This class extends the main Run class with a dict-like API compatible (on a basic level)
    with the legacy neptune-client package.

    Example:

        from neptune_scale.legacy import Run

        run = Run(...)
        run['foo'] = 1
        run['metrics/loss'].append(0.5, step=10)

        run.close()
    """

    def __getitem__(self, key: str) -> Attribute:
        return self._attr_store[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self._attr_store[key] = value



import re
from typing import (
    Optional,
    cast,
)

from neptune_scale.api.validation import verify_type
from neptune_scale.net import projects
from neptune_scale.net.projects import ProjectVisibility

PROJECT_QUALIFIED_NAME_RE = re.compile(r"^((?P<workspace>[\w\-.]+)/)?(?P<project>[\w\-.]+)$")


def create_project(
    name: str,
    *,
    workspace: Optional[str] = None,
    visibility: str = ProjectVisibility.PRIVATE.value,
    description: Optional[str] = None,
    key: Optional[str] = None,
    fail_if_exists: bool = False,
    api_token: Optional[str] = None,
) -> str:
    """Creates a new project in a Neptune workspace.

    Args:
        name (str): Name of the project. Can contain letters and hyphens (-). For example, "project-x".
        workspace (str, optional): Name of your Neptune workspace.
            You can omit this argument if you include the workspace name in the `name` argument.
        visibility: Level of privacy for the project. Options:
            - "pub": Public. Anyone on the internet can see it.
            - "priv" (default): Private. Only users specifically assigned to the project can access it. Requires a plan with
                project-level access control.
            - "workspace" (team workspaces only): Accessible to all workspace members.
        description: Project description. If None, it's left empty.
        key: Project identifier. Must contain 1-10 upper case letters or numbers (at least one letter).
            For example, "PX2". If you leave it out, Neptune generates a project key for you.
        fail_if_exists: If the project already exists and this flag is set to `True`, an error is raised.
        api_token: Account's API token.
            If not provided, the value of the NEPTUNE_API_TOKEN environment variable is used (recommended).

    Returns:
        The name of the new project created.
    """

    verify_type("name", name, str)
    verify_type("workspace", workspace, (str, type(None)))
    verify_type("visibility", visibility, str)
    verify_type("description", description, (str, type(None)))
    verify_type("key", key, (str, type(None)))
    verify_type("api_token", api_token, (str, type(None)))

    workspace, name = extract_workspace_and_project(name=name, workspace=workspace)
    projects.create_project(
        workspace=workspace,
        name=name,
        visibility=visibility,
        description=description,
        key=key,
        fail_if_exists=fail_if_exists,
        api_token=api_token,
    )

    return normalize_project_name(name, workspace)


def extract_workspace_and_project(name: str, workspace: Optional[str] = None) -> tuple[str, str]:
    """Return a tuple of (workspace name, project name) from the provided
    fully qualified project name, or a name + workspace

    >>> extract_workspace_and_project("my-own.workspace_/pr_oj-ect")
    ('my-own.workspace_', 'pr_oj-ect')
    >>> extract_workspace_and_project("project", "workspace")
    ('workspace', 'project')
    >>> extract_workspace_and_project("workspace/project", "workspace")
    ('workspace', 'project')
    >>> extract_workspace_and_project("workspace/project", "another_workspace")
    Traceback (most recent call last):
    ...
    ValueError: The provided `workspace` argument `another_workspace` is different ...
    >>> extract_workspace_and_project("project")
    Traceback (most recent call last):
    ...
    ValueError: Workspace not provided ...
    >>> extract_workspace_and_project("workspace/project!@#")
    Traceback (most recent call last):
    ...
    ValueError: Invalid project name ...
    """
    project_spec = PROJECT_QUALIFIED_NAME_RE.search(name)

    if not project_spec:
        raise ValueError(f"Invalid project name `{name}`")

    extracted_workspace, extracted_project_name = (
        project_spec["workspace"],
        project_spec["project"],
    )

    if not workspace and not extracted_workspace:
        raise ValueError("Workspace not provided in neither project name or the `workspace` parameter.")

    if workspace and extracted_workspace and workspace != extracted_workspace:
        raise ValueError(
            f"The provided `workspace` argument `{workspace}` is different from the one in project name `{name}`"
        )

    final_workspace_name = cast(str, extracted_workspace or workspace)

    return final_workspace_name, extracted_project_name


def normalize_project_name(name: str, workspace: Optional[str] = None) -> str:
    extracted_workspace_name, extracted_project_name = extract_workspace_and_project(name=name, workspace=workspace)

    return f"{extracted_workspace_name}/{extracted_project_name}"


def list_projects(*, api_token: Optional[str] = None) -> list[str]:
    """Lists projects that the account has access to.

    Args:
        api_token: Account's API token.
            If None, the value of the NEPTUNE_API_TOKEN environment variable is used.
            Note: To keep your token secure, use the NEPTUNE_API_TOKEN environment variable rather than placing your
            API token in plain text in your source code.

    Returns:
        List of project names in the form "workspace-name/project-name".
    """

    verify_type("api_token", api_token, (str, type(None)))

    return [
        normalize_project_name(project["name"], project["organizationName"])
        for project in projects.get_project_list(api_token=api_token)
    ]



import multiprocessing
import os
import queue
import threading
from collections.abc import Callable
from functools import partial
from multiprocessing.connection import Connection
from typing import (
    Any,
    Optional,
)

from neptune_scale.util import (
    Daemon,
    get_logger,
)

POLL_TIMEOUT = 0.1

logger = get_logger()


class ProcessLink:
    """Used for bidirectional monitoring and control channel between two processes.

    When a linked process terminates, the `on_link_closed` callback provided in `start()` will be called.
    This way both the parent and child process know when any of them terminated.

    Note that:
     * you MUST call `start()` on the link in both parent and child processes
     * the callback will be called from a thread

    Typically, you would use it like this:
        import time

        # Child process worker
        def worker(link):
            # Required for the link to work
            link.start()
            # Do some work
            time.sleep(1)

        # Parent process
        link = ProcessLink()
        child = multiprocessing.Process(target=worker, args=(link,))
        child.start()

        link.start(on_link_closed=lambda l: print('child closed'))
        link.join() # Wait until the other side is terminated

    See tests/unit/test_process_link.py for more examples.

    The main idea behind this is to create pipes shared between child and parent processes.
    When any process terminates, the other one will receive errors when reading the pipe.

    We could use the `signal` module (registering for SIGCHLD), but this approach has several problems:
     - registering for SIGCHLD is not available on Windows
     - registering for SIGCHLD is brittle, because users might register their own signal handlers, which would either
       break our library, or user code
     - the child process has no way to detect if the parent process dies
    """

    def __init__(self) -> None:
        self._pipes = multiprocessing.Pipe()

        self._on_link_closed: Optional[Callable[[ProcessLink], Any]] = None
        self._on_message_received: Optional[Callable[[ProcessLink, Any], Any]] = None

        # PID of the process that created this object, which should be the parent process.
        # Used to detect if we're calling start() from child or parent.
        self._parent_pid = os.getpid()

        # These fields below are initialized in the start() method, as they cannot be pickled
        # and thus passed to a multiprocessing.Process()

        self._worker: Optional[ProcessLinkWorker] = None
        # Signal end of the startup procedure
        self._start_done: Optional[threading.Event] = None
        self._msg_queue: Optional[queue.SimpleQueue] = None

    def start(
        self,
        *,
        timeout: float = 5.0,
        on_link_closed: Optional[Callable[["ProcessLink"], Any]] = None,
        on_message_received: Optional[Callable[["ProcessLink", Any], Any]] = None,
    ) -> bool:
        """Start the monitoring thread and wait for `timeout` seconds for the other side to start.

        Returns True if the other side is started successfully within the given timeout, False otherwise.

        You can provide callbacks for when the link is closed or when a message is received. The callbacks should
        accept the `ProcessLink` instance as the first argument.

        Note that during this call the `on_terminate` callback can already be called, for example
        if the process fails to start properly.

        This method is NOT thread-safe. It should be called only once, preferably from the main thread.
        """

        if self._worker:
            raise RuntimeError("start() can only be called once")

        self._on_link_closed = on_link_closed
        self._on_message_received = on_message_received
        self._msg_queue = queue.SimpleQueue()

        # Close the "other" end of the pipe depending on which process we're in
        if os.getpid() == self._parent_pid:
            self._pipes[1].close()
            conn = self._pipes[0]
        else:
            self._pipes[0].close()
            conn = self._pipes[1]

        del self._pipes
        self._start_done = threading.Event()

        self._worker = ProcessLinkWorker(
            conn,
            self._start_done,
            start_timeout=timeout,
            msg_queue=self._msg_queue,
            on_link_closed=partial(self._on_link_closed, self) if self._on_link_closed else None,
            on_message_received=partial(self._on_message_received, self) if self._on_message_received else None,
        )
        self._worker.start()

        return self._start_done.wait(timeout=timeout) and not self._worker.is_link_closed

    def stop(self, wait: bool = True, timeout: Optional[float] = None) -> None:
        """Stop monitoring the link. The other side will be notified. The `on_link_closed` callback will NOT be called
        in the calling process. You can optionally wait for the worker to finish.
        """

        if not self._worker:
            raise RuntimeError("You must call start() before calling stop()")

        self._worker.stop_link()
        if wait:
            self.join(timeout)

    def join(self, timeout: Optional[float] = None) -> None:
        """
        Wait for the link monitor to finish. This happens only after the other side is terminated or stop() is called.
        """

        if not self._worker:
            raise RuntimeError("You must call start() before calling join()")

        self._worker.join(timeout=timeout)

    def send(self, message: Any) -> None:
        if not self._msg_queue:
            raise RuntimeError("You must call start() before sending messages")

        self._msg_queue.put(message)


class ProcessLinkWorker(Daemon):
    def __init__(
        self,
        conn: Connection,
        start_done_event: threading.Event,
        start_timeout: float,
        msg_queue: queue.SimpleQueue,
        *,
        on_link_closed: Optional[Callable[[], Any]] = None,
        on_message_received: Optional[Callable[[Any], Any]] = None,
    ):
        super().__init__(sleep_time=0, name="ProcessLink")

        # Indicates if the other side is alive
        self._closed = False

        self._conn = conn
        self._start_timeout = start_timeout
        self._msg_queue = msg_queue
        self._on_link_closed = on_link_closed
        self._on_message_received = on_message_received

        # Signal end of the startup procedure
        self._start_done = start_done_event
        # Set if we've started stopping the link
        self._stop_event = threading.Event()
        self._lock = threading.RLock()

    @property
    def is_link_closed(self) -> bool:
        with self._lock:
            return self._closed

    def run(self) -> None:
        # Before entering the main loop we send and receive a single message.
        # This allows us to make sure that the process on the other end of the pipe is started successfully.
        success = False
        try:
            self._conn.send("started")

            # Give the child process some time for startup
            if not self._conn.poll(timeout=self._start_timeout):
                raise RuntimeError(f"Other side failed to start within {self._start_timeout} seconds")

            if (msg := self._conn.recv()) != "started":
                raise RuntimeError(f"Invalid challenge response: {msg}")
            success = True
        except Exception as e:
            logger.error(f"{self}: unexpected error while sending data: {e}")
            with self._lock:
                self._closed = True
            self.interrupt()

        self._start_done.set()

        # ProcessLink is guaranteed not to call on_link_close after stop() has been called so make sure it is so
        if not success and not self._stop_event.is_set():
            self._safe_callback(self._on_link_closed)
        else:
            super().run()

    def work(self) -> None:
        if self._stop_event.is_set():
            # This will notify the other side about us closing the link
            if not self._conn.closed:
                self._conn.close()

            self.interrupt()
            return

        try:
            self._send_queued_messages()

            # Don't block on self._conn.recv() indefinitely, so we can react to stopping the link
            if not self._conn.poll(timeout=POLL_TIMEOUT) or not self.is_running():
                return

            msg = self._conn.recv()
            if not self._stop_event.is_set():
                self._safe_callback(self._on_message_received, msg)
        except Exception as e:
            # EOFError is expected when the other side closes the connection.
            if not isinstance(e, EOFError):
                logger.error(f"{self}: unexpected error while receiving data: {e}")

            with self._lock:
                self._closed = True

            if not self._stop_event.is_set():
                self._safe_callback(self._on_link_closed)
                self._conn.close()

            self.interrupt()

    def stop_link(self) -> None:
        self._stop_event.set()

    def _safe_callback(self, func: Optional[Callable[..., Any]], *args: Any, **kwargs: Any) -> None:
        if func is None:
            return

        try:
            func(*args, **kwargs)
        except Exception as e:
            logger.error(f"{self}: exception while calling user callback {func}: {e}")

    def _send_queued_messages(self) -> None:
        while True:
            try:
                out = self._msg_queue.get_nowait()
                self._conn.send(out)
            except queue.Empty:
                return



import functools
import itertools
import warnings
from collections.abc import (
    Collection,
    Iterator,
)
from datetime import datetime
from typing import (
    Any,
    Callable,
    Optional,
    Union,
    cast,
)

from neptune_scale.sync.metadata_splitter import MetadataSplitter
from neptune_scale.sync.operations_queue import OperationsQueue

__all__ = ("Attribute", "AttributeStore")


def warn_unsupported_params(fn: Callable) -> Callable:
    # Perform some simple heuristics to detect if a method is called with parameters
    # that are not supported by Scale
    warn = functools.partial(warnings.warn, stacklevel=3)

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):  # type: ignore
        if kwargs.get("wait") is not None:
            warn("The `wait` parameter is not yet implemented and will be ignored.")

        extra_kwargs = set(kwargs.keys()) - {"wait", "step", "timestamp", "steps", "timestamps"}
        if extra_kwargs:
            warn(
                f"`{fn.__name__}()` was called with additional keyword argument(s): `{', '.join(extra_kwargs)}`. "
                "These arguments are not supported by Neptune Scale and will be ignored."
            )

        return fn(*args, **kwargs)

    return wrapper


# TODO: proper typehinting
# AtomType = Union[float, bool, int, str, datetime, list, set, tuple]
ValueType = Any  # Union[float, int, str, bool, datetime, Tuple, List, Dict, Set]


class AttributeStore:
    """
    Responsible for managing local attribute store, and pushing log() operations
    to the provided OperationsQueue -- assuming that there is something on the other
    end consuming the queue (which would be SyncProcess).
    """

    def __init__(self, project: str, run_id: str, operations_queue: OperationsQueue) -> None:
        self._project = project
        self._run_id = run_id
        self._operations_queue = operations_queue
        self._attributes: dict[str, Attribute] = {}

    def __getitem__(self, path: str) -> "Attribute":
        path = cleanup_path(path)
        attr = self._attributes.get(path)
        if attr is None:
            attr = Attribute(self, path)
            self._attributes[path] = attr

        return attr

    def __setitem__(self, key: str, value: ValueType) -> None:
        # TODO: validate type if attr is already known?
        attr = self[key]
        attr.assign(value)

    def log(
        self,
        step: Optional[Union[float, int]] = None,
        timestamp: Optional[Union[datetime, float]] = None,
        configs: Optional[dict[str, Union[float, bool, int, str, datetime, list, set, tuple]]] = None,
        metrics: Optional[dict[str, Union[float, int]]] = None,
        tags_add: Optional[dict[str, Union[list[str], set[str], tuple[str]]]] = None,
        tags_remove: Optional[dict[str, Union[list[str], set[str], tuple[str]]]] = None,
    ) -> None:
        if timestamp is None:
            timestamp = datetime.now()
        elif isinstance(timestamp, float):
            timestamp = datetime.fromtimestamp(timestamp)

        splitter: MetadataSplitter = MetadataSplitter(
            project=self._project,
            run_id=self._run_id,
            step=step,
            timestamp=timestamp,
            configs=configs,
            metrics=metrics,
            add_tags=tags_add,
            remove_tags=tags_remove,
        )

        for operation, metadata_size in splitter:
            self._operations_queue.enqueue(operation=operation, size=metadata_size, key=step)


class Attribute:
    """Objects of this class are returned on dict-like access to Run. Attributes have a path and
    allow logging values under it. Example:

        run = Run(...)
        run['foo'] = 1
        run['nested'] = {'foo': 1, {'bar': {'baz': 2}}}
        run['bar'].append(1, step=10)
    """

    def __init__(self, store: AttributeStore, path: str) -> None:
        self._store = store
        self._path = path

    # TODO: typehint value properly
    @warn_unsupported_params
    def assign(self, value: Any, *, wait: bool = False) -> None:
        data = accumulate_dict_values(value, self._path)
        self._store.log(configs=data)

    @warn_unsupported_params
    def append(
        self,
        value: Union[dict[str, Any], float],
        *,
        step: Union[float, int],
        timestamp: Optional[Union[float, datetime]] = None,
        wait: bool = False,
        **kwargs: Any,
    ) -> None:
        data = accumulate_dict_values(value, self._path)
        self._store.log(metrics=data, step=step, timestamp=timestamp)

    @warn_unsupported_params
    # TODO: this should be Iterable in Run as well
    # def add(self, values: Union[str, Iterable[str]], *, wait: bool = False) -> None:
    def add(self, values: Union[str, Union[list[str], set[str], tuple[str]]], *, wait: bool = False) -> None:
        if isinstance(values, str):
            values = (values,)
        self._store.log(tags_add={self._path: values})

    @warn_unsupported_params
    # TODO: this should be Iterable in Run as well
    # def remove(self, values: Union[str, Iterable[str]], *, wait: bool = False) -> None:
    def remove(self, values: Union[str, Union[list[str], set[str], tuple[str]]], *, wait: bool = False) -> None:
        if isinstance(values, str):
            values = (values,)
        self._store.log(tags_remove={self._path: values})

    @warn_unsupported_params
    def extend(
        self,
        values: Collection[Union[float, int]],
        *,
        steps: Collection[Union[float, int]],
        timestamps: Optional[Collection[Union[float, datetime]]] = None,
        wait: bool = False,
        **kwargs: Any,
    ) -> None:
        # TODO: make this compatible with the old client
        assert len(values) == len(steps)

        if timestamps is not None:
            assert len(timestamps) == len(values)
        else:
            timestamps = cast(tuple, itertools.repeat(datetime.now()))

        for value, step, timestamp in zip(values, steps, timestamps):
            self.append(value, step=step, timestamp=timestamp, wait=wait)

    # TODO: add value type validation to all the methods
    # TODO: change Run API to typehint timestamp as Union[datetime, float]


def cleanup_path(path: str) -> str:
    """
    >>> cleanup_path('/a/b/c')
    'a/b/c'
    >>> cleanup_path('a/b/c/')
    Traceback (most recent call last):
    ...
    ValueError: Invalid path: `a/b/c/`. Path must not end with a slash.
    >>> cleanup_path('a//b/c')
    Traceback (most recent call last):
    ...
    ValueError: Invalid path: `a//b/c`. Path components must not be empty.
    >>> cleanup_path('a/ /b/c')
    Traceback (most recent call last):
    ...
    ValueError: Invalid path: `a/ /b/c`. Path components cannot contain leading or trailing whitespace.
    >>> cleanup_path('a/b/c ')
    Traceback (most recent call last):
    ...
    ValueError: Invalid path: `a/b/c `. Path components cannot contain leading or trailing whitespace.
    """

    if path.strip() in ("", "/"):
        raise ValueError(f"Invalid path: `{path}`.")

    orig_parts = path.split("/")
    parts = [x.strip() for x in orig_parts]

    for i, part in enumerate(parts):
        if part != orig_parts[i]:
            raise ValueError(f"Invalid path: `{path}`. Path components cannot contain leading or trailing whitespace.")

    # Skip the first slash, if present
    if parts[0] == "":
        parts = parts[1:]

    if parts[-1] == "":
        raise ValueError(f"Invalid path: `{path}`. Path must not end with a slash.")

    if not all(parts):
        raise ValueError(f"Invalid path: `{path}`. Path components must not be empty.")

    return "/".join(parts)


def accumulate_dict_values(value: Union[ValueType, dict[str, ValueType]], path_or_base: str) -> dict[str, Any]:
    """
    If value is a dict, flatten nested dictionaries into a single dict with unwrapped paths, each
    starting with `path_or_base`.

    If value is an atom, return a dict with a single entry `path_or_base` -> `value`.

    >>> accumulate_dict_values(1, "foo")
    {'foo': 1}
    >>> accumulate_dict_values({"bar": 1, 'l0/l1': 2, 'l3':{"l4": 3}}, "foo")
    {'foo/bar': 1, 'foo/l0/l1': 2, 'foo/l3/l4': 3}
    """

    if isinstance(value, dict):
        data = {"/".join(path): value for path, value in iter_nested(value, path_or_base)}
    else:
        data = {path_or_base: value}

    return data


def iter_nested(dict_: dict[str, ValueType], path: str) -> Iterator[tuple[tuple[str, ...], ValueType]]:
    """Iterate a nested dictionary, yielding a tuple of path components and value.

    >>> list(iter_nested({"foo": 1, "bar": {"baz": 2}}, "base"))
    [(('base', 'foo'), 1), (('base', 'bar', 'baz'), 2)]
    >>> list(iter_nested({"foo":{"bar": 1}, "bar":{"baz": 2}}, "base"))
    [(('base', 'foo', 'bar'), 1), (('base', 'bar', 'baz'), 2)]
    >>> list(iter_nested({"foo": 1, "bar": 2}, "base"))
    [(('base', 'foo'), 1), (('base', 'bar'), 2)]
    >>> list(iter_nested({"foo": {}}, ""))
    Traceback (most recent call last):
    ...
    ValueError: The dictionary cannot be empty or contain empty nested dictionaries.
    """

    parts = tuple(path.split("/"))
    yield from _iter_nested(dict_, parts)


def _iter_nested(dict_: dict[str, ValueType], path_acc: tuple[str, ...]) -> Iterator[tuple[tuple[str, ...], ValueType]]:
    if not dict_:
        raise ValueError("The dictionary cannot be empty or contain empty nested dictionaries.")

    for key, value in dict_.items():
        current_path = path_acc + (key,)
        if isinstance(value, dict):
            yield from _iter_nested(value, current_path)
        else:
            yield current_path, value



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

__all__ = ("HostedApiClient", "MockedApiClient", "ApiClient", "backend_factory", "with_api_errors_handling")

import abc
import functools
import os
import uuid
from collections.abc import Callable
from dataclasses import dataclass
from http import HTTPStatus
from typing import (
    Any,
    Literal,
)

import httpx
from httpx import Timeout
from neptune_api import (
    AuthenticatedClient,
    Client,
)
from neptune_api.api.backend import get_client_config
from neptune_api.api.data_ingestion import (
    check_request_status_bulk,
    submit_operation,
)
from neptune_api.auth_helpers import exchange_api_key
from neptune_api.credentials import Credentials
from neptune_api.errors import (
    InvalidApiTokenException,
    UnableToDeserializeApiKeyError,
    UnableToExchangeApiKeyError,
    UnableToRefreshTokenError,
    UnexpectedStatus,
)
from neptune_api.models import (
    ClientConfig,
    Error,
)
from neptune_api.proto.google_rpc.code_pb2 import Code
from neptune_api.proto.neptune_pb.ingest.v1.ingest_pb2 import IngestCode
from neptune_api.proto.neptune_pb.ingest.v1.pub.client_pb2 import (
    BulkRequestStatus,
    RequestId,
    RequestIdList,
    SubmitResponse,
)
from neptune_api.proto.neptune_pb.ingest.v1.pub.ingest_pb2 import RunOperation
from neptune_api.proto.neptune_pb.ingest.v1.pub.request_status_pb2 import RequestStatus
from neptune_api.types import Response

from neptune_scale.exceptions import (
    NeptuneConnectionLostError,
    NeptuneInvalidCredentialsError,
    NeptuneUnableToAuthenticateError,
)
from neptune_scale.sync.parameters import REQUEST_TIMEOUT
from neptune_scale.util.abstract import Resource
from neptune_scale.util.envs import ALLOW_SELF_SIGNED_CERTIFICATE
from neptune_scale.util.logger import get_logger

logger = get_logger()


@dataclass
class TokenRefreshingURLs:
    authorization_endpoint: str
    token_endpoint: str

    @classmethod
    def from_dict(cls, data: dict) -> TokenRefreshingURLs:
        return TokenRefreshingURLs(
            authorization_endpoint=data["authorization_endpoint"], token_endpoint=data["token_endpoint"]
        )


def get_config_and_token_urls(
    *, credentials: Credentials, verify_ssl: bool
) -> tuple[ClientConfig, TokenRefreshingURLs]:
    with Client(
        base_url=credentials.base_url,
        follow_redirects=True,
        verify_ssl=verify_ssl,
        timeout=Timeout(timeout=REQUEST_TIMEOUT),
    ) as client:
        config = get_client_config.sync(client=client)
        if config is None or isinstance(config, Error):
            raise RuntimeError(f"Failed to get client config: {config}")
        response = client.get_httpx_client().get(config.security.open_id_discovery)
        token_urls = TokenRefreshingURLs.from_dict(response.json())
    return config, token_urls


def create_auth_api_client(
    *, credentials: Credentials, config: ClientConfig, token_refreshing_urls: TokenRefreshingURLs, verify_ssl: bool
) -> AuthenticatedClient:
    return AuthenticatedClient(
        base_url=credentials.base_url,
        credentials=credentials,
        client_id=config.security.client_id,
        token_refreshing_endpoint=token_refreshing_urls.token_endpoint,
        api_key_exchange_callback=exchange_api_key,
        follow_redirects=True,
        verify_ssl=verify_ssl,
        timeout=Timeout(timeout=REQUEST_TIMEOUT),
    )


class ApiClient(Resource, abc.ABC):
    @abc.abstractmethod
    def submit(self, operation: RunOperation, family: str) -> Response[SubmitResponse]: ...

    @abc.abstractmethod
    def check_batch(self, request_ids: list[str], project: str) -> Response[BulkRequestStatus]: ...


class HostedApiClient(ApiClient):
    def __init__(self, api_token: str) -> None:
        credentials = Credentials.from_api_key(api_key=api_token)

        verify_ssl: bool = os.environ.get(ALLOW_SELF_SIGNED_CERTIFICATE, "False").lower() in ("false", "0")

        logger.debug("Trying to connect to Neptune API")
        config, token_urls = get_config_and_token_urls(credentials=credentials, verify_ssl=verify_ssl)
        self.backend = create_auth_api_client(
            credentials=credentials, config=config, token_refreshing_urls=token_urls, verify_ssl=verify_ssl
        )
        logger.debug("Connected to Neptune API")

    def submit(self, operation: RunOperation, family: str) -> Response[SubmitResponse]:
        return submit_operation.sync_detailed(client=self.backend, body=operation, family=family)

    def check_batch(self, request_ids: list[str], project: str) -> Response[BulkRequestStatus]:
        return check_request_status_bulk.sync_detailed(
            client=self.backend,
            project_identifier=project,
            body=RequestIdList(ids=[RequestId(value=request_id) for request_id in request_ids]),
        )

    def close(self) -> None:
        logger.debug("Closing API client")
        self.backend.__exit__()


class MockedApiClient(ApiClient):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        pass

    def submit(self, operation: RunOperation, family: str) -> Response[SubmitResponse]:
        operation_count = len(operation.update_batch.snapshots) if operation.update_batch.snapshots else 1
        request_ids = [str(uuid.uuid4()) for _ in range(operation_count)]
        response = SubmitResponse(request_id=request_ids[-1], request_ids=request_ids)
        return Response(content=b"", parsed=response, status_code=HTTPStatus.OK, headers={})

    def check_batch(self, request_ids: list[str], project: str) -> Response[BulkRequestStatus]:
        response_body = BulkRequestStatus(
            statuses=list(
                map(
                    lambda _: RequestStatus(
                        code_by_count=[RequestStatus.CodeByCount(count=1, code=Code.OK, detail=IngestCode.OK)]
                    ),
                    request_ids,
                )
            )
        )
        return Response(content=b"", parsed=response_body, status_code=HTTPStatus.OK, headers={})


def backend_factory(api_token: str, mode: Literal["async", "disabled"]) -> ApiClient:
    if mode == "disabled":
        return MockedApiClient()
    return HostedApiClient(api_token=api_token)


def with_api_errors_handling(func: Callable[..., Any]) -> Callable[..., Any]:
    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            return func(*args, **kwargs)
        except (InvalidApiTokenException, UnableToDeserializeApiKeyError):
            raise NeptuneInvalidCredentialsError()
        except (UnableToRefreshTokenError, UnableToExchangeApiKeyError, UnexpectedStatus):
            raise NeptuneUnableToAuthenticateError()
        except httpx.RequestError:
            raise NeptuneConnectionLostError()
        except Exception as e:
            raise e

    return wrapper
